(* $Id$ *)


(**
   Numerical Non-Relational Abstract Domains.

   A numerical non-relational abstract domain is an abstract lattice that is in
   Galois connection with the concrete lattice (℘(ℤ), ⊆).  The domain comes with
   abstract numerical operations.

   {L {b Implemented by:} {! DomConstant}, {! DomSign}.}
 *)


(**
   Common interface to all numerical non-relational abstract domains.
 *)
module type S =
sig
  (**
     Underlying lattice (A, ⊑).
   *)
  include Lattice.S

  (**
     Abstraction function α (for singletons).
   *)
  val abs : int -> t

  (**
     [empty a] tests whether the concretization of [a] is the empty set.
   *)
  val empty : t -> bool

  (**
     Abstract numerical operations.
   *)
  module Op :
  sig
    (**
       [add a b] returns the abstraction of
       \{x+y | x ∈ γ([a]) ∧ y ∈ γ([b])\}.
     *)
    val add : t -> t -> t

    (**
       [sub a b] returns the abstraction of
       \{x-y | x ∈ γ([a]) ∧ y ∈ γ([b])\}.
     *)
    val sub : t -> t -> t

    (**
       [mul a b] returns the abstraction of
       \{x*y | x ∈ γ([a]) ∧ y ∈ γ([b])\}.
     *)
    val mul : t -> t -> t

    (**
       [div a b] returns the abstraction of
       \{x/y | x ∈ γ([a]) ∧ y ∈ γ([b]) ∧ y≠0\}.
     *)
    val div : t -> t -> t

    (**
       [equality a b] tests whether there exist
       x ∈ γ([a]) and y ∈ γ([b]) such that x=y.
     *)
    val equality : t -> t -> bool

    (**
       [inequality a b] tests whether there exist
       x ∈ γ([a]) and y ∈ γ([b]) such that x≤y.
     *)
    val inequality : t -> t -> bool
  end
end
