(* $Id$ *)


(*
   Unit tests for the DomConstant module.
 *)


open TestCore
open DomConstant

(* Import utility functions. *)
module TU = TestUtil.OfNumericalDomain (DomConstant)
open TU

module Make () =
struct
(* Define some elements of the constant domain (except bot and top). *)
let minus_five = abs (-5)
let minus_three = abs (-3)
let zero = abs 0
let plus_three = abs 3
let plus_five = abs 5
let elements =
  [bot ; minus_five ; minus_three ; zero ; plus_three ; plus_five ; top]

let test_print () =
  assert_print ~msg:"print bot"         "⊥"  print bot ;
  assert_print ~msg:"print minus_five"  "-5" print minus_five ;
  assert_print ~msg:"print minus_three" "-3" print minus_three ;
  assert_print ~msg:"print zero"        "0"  print zero ;
  assert_print ~msg:"print plus_three"  "3"  print plus_three ;
  assert_print ~msg:"print plus_five"   "5"  print plus_five ;
  assert_print ~msg:"print top"         "⊤"  print top

let test_abs () =
  assert_equal_t ~msg:"abs -13" "-13" (abs (-13)) ;
  assert_equal_t ~msg:"abs 0"   "0"   (abs 0) ;
  assert_equal_t ~msg:"abs 7"   "7"   (abs 7)

let test_empty () =
  assert_bool ~msg:"empty bot"        true  (empty bot) ;
  assert_bool ~msg:"empty minus_five" false (empty minus_five) ;
  assert_bool ~msg:"empty zero"       false (empty zero) ;
  assert_bool ~msg:"empty plus_three" false (empty plus_three) ;
  assert_bool ~msg:"empty top"        false (empty top)

let test_add () =
  let aux a b expected =
    assert_equal_t ~msg:(message "add" a b) expected (Op.add a b) ;
    if not (a == b) then
      assert_equal_t ~msg:(message "add" b a) expected (Op.add b a)
  in
  List.iter (fun a -> aux bot a "⊥") elements ;
  aux minus_five minus_five "-10" ;
  aux minus_five plus_three "-2" ;
  aux zero minus_five "-5" ;
  aux zero zero "0" ;
  aux zero plus_three "3" ;
  aux zero top "⊤" ;
  aux plus_three plus_three "6" ;
  aux top minus_five "⊤" ;
  aux top plus_three "⊤" ;
  aux top top "⊤"

let opposite a =
  match a with
  | "⊥" | "⊤" | "0" -> a
  | "-8" -> "8"
  | "-5" -> "5"
  | "-3" -> "3"
  | "3" -> "-3"
  | "5" -> "-5"
  | "8" -> "-8"
  | _ -> invalid_arg "Test_DomConstant.opposite"

let test_sub () =
  let aux a b expected =
    assert_equal_t ~msg:(message "sub" a b) expected (Op.sub a b) ;
    if not (a == b) then
      assert_equal_t ~msg:(message "sub" b a) (opposite expected) (Op.sub b a)
  in
  List.iter (fun a -> aux bot a "⊥") elements ;
  aux minus_five minus_five "0" ;
  aux minus_five plus_three "-8" ;
  aux zero minus_five "5" ;
  aux zero zero "0" ;
  aux zero plus_three "-3" ;
  aux zero top "⊤" ;
  aux plus_three plus_three "0" ;
  aux top minus_five "⊤" ;
  aux top plus_three "⊤" ;
  aux top top "⊤"

let test_mul () =
  let aux a b expected =
    assert_equal_t ~msg:(message "mul" a b) expected (Op.mul a b) ;
    if not (a == b) then
      assert_equal_t ~msg:(message "mul" b a) expected (Op.mul b a)
  in
  List.iter (fun a -> aux bot a "⊥") elements ;
  aux minus_five minus_five "25" ;
  aux minus_five plus_three "-15" ;
  aux zero minus_five "0" ;
  aux zero zero "0" ;
  aux zero plus_three "0" ;
  aux zero top "0" ;
  aux plus_three plus_three "9" ;
  aux top minus_five "⊤" ;
  aux top plus_three "⊤" ;
  aux top top "⊤"

let test_div () =
  let aux a b expected =
    assert_equal_t ~msg:(message "div" a b) expected (Op.div a b)
  in
  List.iter (fun a -> aux bot a "⊥") elements ;
  aux minus_five minus_five "1" ;
  aux minus_five zero "⊥" ;
  aux minus_five plus_three "-1" ;
  aux minus_five top "⊤" ;
  aux (abs (-13)) plus_three "-4" ;
  aux zero minus_five "0" ;
  aux zero zero "⊥" ;
  aux zero plus_three "0" ;
  aux zero top "0" ;
  aux plus_three minus_five "0" ;
  aux plus_three zero "⊥" ;
  aux plus_three plus_three "1" ;
  aux plus_three top "⊤" ;
  aux (abs (17)) minus_five "-3" ;
  aux top minus_five "⊤" ;
  aux top zero "⊥" ;
  aux top plus_three "⊤" ;
  aux top top "⊤"

let test_equality () =
  let aux a b expected =
    assert_bool ~msg:(message "equality" a b) expected (Op.equality a b) ;
    if not (a == b) then
      assert_bool ~msg:(message "equality" b a) expected (Op.equality b a)
  in
  List.iter (fun a -> aux bot a false) elements ;
  aux minus_five minus_five true ;
  aux minus_five plus_three false ;
  aux zero minus_five false ;
  aux zero zero true ;
  aux zero plus_three false ;
  aux zero top true ;
  aux plus_three plus_three true ;
  aux top minus_five true ;
  aux top plus_three true ;
  aux top top true

let test_inequality () =
  let opposite a = Op.sub zero a
  in
  let aux a b expected =
    assert_bool ~msg:(message "inequality" a b) expected (Op.inequality a b) ;
    if not (a == opposite b) then
      let a' = opposite a
      and b' = opposite b
      and expected' = expected
      in
      assert_bool ~msg:(message "inequality" b' a') expected' (Op.inequality b' a')
  in
  List.iter (fun a -> aux bot a false) elements ;
  aux minus_five minus_five true ;
  aux minus_five plus_three true ;
  aux zero minus_five false ;
  aux zero zero true ;
  aux zero plus_three true ;
  aux zero top true ;
  aux plus_three minus_five false ;
  aux plus_three plus_three true ;
  aux top minus_five true ;
  aux top plus_three true ;
  aux top top true

(* Create the generic lattice test cases. *)
module M =
struct
  module L = DomConstant
  let elements = elements
  let relation =
    [
      (bot, minus_five) ;
      (bot, minus_three) ;
      (bot, zero) ;
      (bot, plus_three) ;
      (bot, plus_five) ;
      (minus_five, top) ;
      (minus_three, top) ;
      (zero, top) ;
      (plus_three, top) ;
      (plus_five, top) ;
    ]
end

module T = TestLattice.Make (M)

(* Collection of all tests. *)
let tests prefix =
  [
    prefix ^ ".print", test_print ;
  ]
  @
  (T.tests prefix)
  @
  [
    prefix ^ ".abs", test_abs ;
    prefix ^ ".empty", test_empty ;
    prefix ^ ".add", test_add ;
    prefix ^ ".sub", test_sub ;
    prefix ^ ".mul", test_mul ;
    prefix ^ ".div", test_div ;
    prefix ^ ".equality", test_equality ;
    prefix ^ ".inequality", test_inequality ;
  ]
end

(* This test suite. *)
let suite = ("DomConstant", lazy (let open Make () in tests "domconstant"))
