(* $Id$ *)


(*
   Unit tests for the DomSign module.
 *)


open TestCore
open DomSign

(* Import utility functions. *)
module TU = TestUtil.OfNumericalDomain (DomSign)
open TU

module Make () =
struct
(* Elements of the sign domain (except bot and top). *)
let negative = abs (-1)
let zero = abs 0
let positive = abs 1
let nonpositive = lub negative zero
let nonzero = lub negative positive
let nonnegative = lub zero positive
let elements =
  [bot ; negative ; zero ; positive ; nonpositive ; nonzero ; nonnegative ; top]

let test_print () =
  assert_print ~msg:"print bot"         "⊥"  print bot ;
  assert_print ~msg:"print negative"    "-"  print negative ;
  assert_print ~msg:"print zero"        "0"  print zero ;
  assert_print ~msg:"print positive"    "+"  print positive ;
  assert_print ~msg:"print nonpositive" "-0" print nonpositive ;
  assert_print ~msg:"print nonzero"     "-+" print nonzero ;
  assert_print ~msg:"print nonnegative" "0+" print nonnegative ;
  assert_print ~msg:"print top"         "⊤"  print top

let test_abs () =
  assert_equal_t ~msg:"abs -13" "-" (abs (-13)) ;
  assert_equal_t ~msg:"abs 0"   "0" (abs 0) ;
  assert_equal_t ~msg:"abs 7"   "+" (abs 7)

let test_empty () =
  assert_bool ~msg:"empty bot"      true  (empty bot) ;
  assert_bool ~msg:"empty negative" false (empty negative) ;
  assert_bool ~msg:"empty zero"     false (empty zero) ;
  assert_bool ~msg:"empty nonzero"  false (empty nonzero) ;
  assert_bool ~msg:"empty top"      false (empty top)

let test_add () =
  let aux a b expected =
    assert_equal_t ~msg:(message "add" a b) expected (Op.add a b) ;
    if not (a == b) then
      assert_equal_t ~msg:(message "add" b a) expected (Op.add b a)
  in
  List.iter (fun a -> aux bot a "⊥") elements ;
  aux negative negative "-" ;
  aux negative positive "⊤" ;
  aux negative nonnegative "⊤" ;
  aux zero zero "0" ;
  aux zero positive "+" ;
  aux zero nonnegative "0+" ;
  aux zero nonzero "-+" ;
  aux zero top "⊤" ;
  aux positive positive "+" ;
  aux positive nonpositive "⊤" ;
  aux nonpositive nonpositive "-0" ;
  aux nonpositive nonnegative "⊤" ;
  aux nonzero negative "⊤" ;
  aux nonzero nonpositive "⊤" ;
  aux nonnegative nonnegative "0+" ;
  aux top negative "⊤" ;
  aux top nonpositive "⊤" ;
  aux top top "⊤"

let opposite a =
  match a with
  | "⊥" | "⊤" | "0" | "-+" -> a
  | "-" -> "+"
  | "+" -> "-"
  | "-0" -> "0+"
  | "0+" -> "-0"
  | _ -> invalid_arg "Test_DomSign.opposite"

let test_sub () =
  let aux a b expected =
    assert_equal_t ~msg:(message "sub" a b) expected (Op.sub a b) ;
    if not (a == b) then
      assert_equal_t ~msg:(message "sub" b a) (opposite expected) (Op.sub b a)
  in
  List.iter (fun a -> aux bot a "⊥") elements ;
  aux negative negative "⊤" ;
  aux negative positive "-" ;
  aux negative nonnegative "-" ;
  aux zero zero "0" ;
  aux zero positive "-" ;
  aux zero nonnegative "-0" ;
  aux zero nonzero "-+" ;
  aux zero top "⊤" ;
  aux positive positive "⊤" ;
  aux positive nonpositive "+" ;
  aux nonpositive nonpositive "⊤" ;
  aux nonpositive nonnegative "-0" ;
  aux nonzero negative "⊤" ;
  aux nonzero nonpositive "⊤" ;
  aux nonnegative nonnegative "⊤" ;
  aux top negative "⊤" ;
  aux top nonpositive "⊤" ;
  aux top top "⊤"

let test_mul () =
  let aux a b expected =
    assert_equal_t ~msg:(message "mul" a b) expected (Op.mul a b) ;
    if not (a == b) then
      assert_equal_t ~msg:(message "mul" b a) expected (Op.mul b a)
  in
  List.iter (fun a -> aux bot a "⊥") elements ;
  aux negative negative "+" ;
  aux negative positive "-" ;
  aux negative nonnegative "-0" ;
  aux zero zero "0" ;
  aux zero positive "0" ;
  aux zero nonnegative "0" ;
  aux zero nonzero "0" ;
  aux zero top "0" ;
  aux positive positive "+" ;
  aux positive nonpositive "-0" ;
  aux nonpositive nonpositive "0+" ;
  aux nonpositive nonnegative "-0" ;
  aux nonzero negative "-+" ;
  aux nonzero nonpositive "⊤" ;
  aux nonnegative nonnegative "0+" ;
  aux top negative "⊤" ;
  aux top nonpositive "⊤" ;
  aux top top "⊤"

let inverse a =
  match a with
  | "⊥" | "⊤" -> a
  | "-" | "-0" -> "-0"
  | "0" -> "⊥"
  | "+" | "0+" -> "0+"
  | "-+" -> "⊤"
  | _ -> invalid_arg "Test_DomSign.inverse"

let test_div () =
  let aux a b expected =
    assert_equal_t ~msg:(message "div" a b) expected (Op.div a b) ;
    if not (a == b) then
      assert_equal_t ~msg:(message "div" b a) (inverse expected) (Op.div b a)
  in
  List.iter (fun a -> aux bot a "⊥") elements ;
  aux negative negative "0+" ;
  aux negative positive "-0" ;
  aux negative nonnegative "-0" ;
  aux zero zero "⊥" ;
  aux zero positive "0" ;
  aux zero nonnegative "0" ;
  aux zero nonzero "0" ;
  aux zero top "0" ;
  aux positive positive "0+" ;
  aux positive nonpositive "-0" ;
  aux nonpositive nonpositive "0+" ;
  aux nonpositive nonnegative "-0" ;
  aux nonzero negative "⊤" ;
  aux nonzero nonpositive "⊤" ;
  aux nonnegative nonnegative "0+" ;
  aux top negative "⊤" ;
  aux top nonpositive "⊤" ;
  aux top top "⊤"

let test_equality () =
  let aux a b expected =
    assert_bool ~msg:(message "equality" a b) expected (Op.equality a b) ;
    if not (a == b) then
      assert_bool ~msg:(message "equality" b a) expected (Op.equality b a)
  in
  List.iter (fun a -> aux bot a false) elements ;
  aux negative negative true ;
  aux negative positive false ;
  aux negative nonnegative false ;
  aux zero negative false ;
  aux zero zero true ;
  aux zero positive false ;
  aux zero nonpositive true ;
  aux zero nonnegative true ;
  aux zero nonzero false ;
  aux zero top true ;
  aux positive positive true ;
  aux positive nonpositive false ;
  aux nonpositive nonpositive true ;
  aux nonpositive nonnegative true ;
  aux nonzero negative true ;
  aux nonzero nonpositive true ;
  aux nonnegative nonnegative true ;
  aux top negative true ;
  aux top nonpositive true ;
  aux top top true

let test_inequality () =
  let opposite a = Op.sub zero a
  in
  let aux a b expected =
    assert_bool ~msg:(message "inequality" a b) expected (Op.inequality a b) ;
    if not (a == opposite b) then
      let a' = opposite a
      and b' = opposite b
      and expected' = expected
      in
      assert_bool ~msg:(message "inequality" b' a') expected' (Op.inequality b' a')
  in
  List.iter (fun a -> aux bot a false) elements ;
  aux negative negative true ;
  aux negative positive true ;
  aux negative nonnegative true ;
  aux zero negative false ;
  aux zero zero true ;
  aux zero positive true ;
  aux zero nonpositive true ;
  aux zero nonnegative true ;
  aux zero nonzero true ;
  aux zero top true ;
  aux positive negative false ;
  aux positive nonpositive false ;
  aux nonpositive nonpositive true ;
  aux nonpositive nonnegative true ;
  aux nonzero negative true ;
  aux nonzero nonpositive true ;
  aux nonnegative nonpositive true ;
  aux top negative true ;
  aux top nonpositive true ;
  aux top top true

(* Create the generic lattice test cases. *)
module M =
struct
  module L = DomSign
  let elements = elements
  let relation =
    [
      (bot, negative) ;
      (bot, zero) ;
      (bot, positive) ;
      (negative, nonpositive) ;
      (negative, nonzero) ;
      (zero, nonpositive) ;
      (zero, nonnegative) ;
      (positive, nonzero) ;
      (positive, nonnegative) ;
      (nonpositive, top) ;
      (nonzero, top) ;
      (nonnegative, top) ;
    ]
end

module T = TestLattice.Make (M)

(* Collection of all tests. *)
let tests prefix =
  [
    prefix ^ ".print", test_print ;
  ]
  @
  (T.tests prefix)
  @
  [
    prefix ^ ".abs", test_abs ;
    prefix ^ ".empty", test_empty ;
    prefix ^ ".add", test_add ;
    prefix ^ ".sub", test_sub ;
    prefix ^ ".mul", test_mul ;
    prefix ^ ".div", test_div ;
    prefix ^ ".equality", test_equality ;
    prefix ^ ".inequality", test_inequality ;
  ]
end

(* This test suite. *)
let suite = ("DomSign", lazy (let open Make () in tests "domsign"))
